# SPDX-FileCopyrightText: 2005 Stani Michiels (Stani)
# SPDX-FileCopyrightText: 2008-2009 Remigiusz Fiedler (migius)
#
# SPDX-License-Identifier: GPL-2.0-or-later

# dxfLibrary.py : provides functions for generating DXF files
# --------------------------------------------------------------------------
__version__ = "v1.35 - 2010.06.23"
__author__ = "Stani Michiels(Stani), Remigiusz Fiedler(migius)"
__license__ = "GPL"
__url__ = "http://wiki.blender.org/index.php/Extensions:2.4/Py/Scripts/Export/DXF"
__bpydoc__ ="""The library to export geometry data to DXF format r12 version.

Copyright %s
Version %s
License %s
Homepage %s

See the homepage for documentation.
Dedicated thread on BlenderArtists: http://blenderartists.org/forum/showthread.php?t=136439

IDEAs:
-

TODO:
- add support for DXFr14 version (needs extended file header)
- add support for DXF-SPLINEs (possible first in DXFr14 version)
- add support for DXF-MTEXT
- add user preset for floating point precision (3-16?)

History
v1.35 - 2010.06.23 by migius
 - added (as default) writing to DXF file without RAM-buffering: faster and low-RAM-machines friendly
v1.34 - 2010.06.20 by migius
 - bugfix POLYFACE
 - added DXF-flags for POLYLINE and VERTEX class (NURBS-export)
v1.33 - 2009.07.03 by migius
 - fix MTEXT newline bug (not supported by DXF-Exporter yet)
 - modif _point(): converts all coords to floats
 - modif LineType class: implement elements
 - added VPORT class, incl. defaults
 - fix Insert class
v1.32 - 2009.06.06 by migius
 - modif Style class: changed defaults to widthFactor=1.0, obliqueAngle=0.0
 - modif Text class: alignment parameter reactivated
v1.31 - 2009.06.02 by migius
 - modif _Entity class: added paperspace,elevation
v1.30 - 2009.05.28 by migius
 - bugfix 3dPOLYLINE/POLYFACE: VERTEX needs x,y,z coordinates, index starts with 1 not 0
v1.29 - 2008.12.28 by Yorik
 - modif POLYLINE to support bulge segments
v1.28 - 2008.12.13 by Steeve/BlenderArtists
 - bugfix for EXTMIN/EXTMAX to suit Cycas-CAD
v1.27 - 2008.10.07 by migius
 - beautifying output code: keys whitespace prefix
 - refactoring DXF-strings format: NewLine moved to the end of
v1.26 - 2008.10.05 by migius
 - modif POLYLINE to support POLYFACE
v1.25 - 2008.09.28 by migius
 - modif FACE class for r12
v1.24 - 2008.09.27 by migius
 - modif POLYLINE class for r12
 - changing output format from r9 to r12(AC1009)
v1.1 (20/6/2005) by www.stani.be/python/sdxf
 - Python library to generate dxf drawings
______________________________________________________________
""" % (__author__,__version__,__license__,__url__)

#import Blender
#from Blender import Mathutils, Window, Scene, sys, Draw
#import BPyMessages

try:
    import copy
    #from struct import pack
except:
    copy = None

####1) Private (only for developers)
_HEADER_POINTS=['insbase','extmin','extmax']

#---helper functions-----------------------------------
def _point(x,index=0):
    """Convert tuple to a dxf point"""
    return '\n'.join([' %s\n%s'%((i+1)*10+index,float(x[i])) for i in range(len(x))])

def _points(plist):
    """Convert a list of tuples to dxf points"""
    out = '\n'.join([_point(plist[i],i)for i in range(len(plist))])
    return out

#---base classes----------------------------------------
class _Call:
    """Makes a callable class."""
    def copy(self):
        """Returns a copy."""
        return copy.deepcopy(self)

    def __call__(self,**attrs):
        """Returns a copy with modified attributes."""
        copied=self.copy()
        for attr in attrs:setattr(copied,attr,attrs[attr])
        return copied

#-------------------------------------------------------
class _Entity(_Call):
    """Base class for _common group codes for entities."""
    def __init__(self,paperspace=None,color=None,layer='0',
                 lineType=None,lineTypeScale=None,lineWeight=None,
                 extrusion=None,elevation=None,thickness=None,
                 parent=None):
        """None values will be omitted."""
        self.paperspace   = paperspace
        self.color        = color
        self.layer        = layer
        self.lineType      = lineType
        self.lineTypeScale  = lineTypeScale
        self.lineWeight  = lineWeight
        self.extrusion    = extrusion
        self.elevation    = elevation
        self.thickness    = thickness
        #self.visible     = visible
        self.parent      = parent

    def _common(self):
        """Return common group codes as a string."""
        if self.parent:parent=self.parent
        else:parent=self
        result =''
        if parent.paperspace==1: result+='  67\n1\n'
        if parent.layer!=None: result+='  8\n%s\n'%parent.layer
        if parent.color!=None: result+=' 62\n%s\n'%parent.color
        if parent.lineType!=None: result+='  6\n%s\n'%parent.lineType
        # TODO: if parent.lineWeight!=None: result+='370\n%s\n'%parent.lineWeight
        # TODO: if parent.visible!=None: result+='60\n%s\n'%parent.visible
        if parent.lineTypeScale!=None: result+=' 48\n%s\n'%parent.lineTypeScale
        if parent.elevation!=None: result+=' 38\n%s\n'%parent.elevation
        if parent.thickness!=None: result+=' 39\n%s\n'%parent.thickness
        if parent.extrusion!=None: result+='%s\n'%_point(parent.extrusion,200)
        return result

#--------------------------
class _Entities:
    """Base class to deal with composed objects."""
    def __dxf__(self):
        return []

    def __str__(self):
        return ''.join([str(x) for x in self.__dxf__()])

#--------------------------
class _Collection(_Call):
    """Base class to expose entities methods to main object."""
    def __init__(self,entities=[]):
        self.entities=copy.copy(entities)
        #link entities methods to drawing
        for attr in dir(self.entities):
            if attr[0]!='_':
                attrObject=getattr(self.entities,attr)
                if callable(attrObject):
                    setattr(self,attr,attrObject)

####2) Constants
#---color values
BYBLOCK=0
BYLAYER=256

#---block-type flags (bit coded values, may be combined):
ANONYMOUS =1  # This is an anonymous block generated by hatching, associative dimensioning, other internal operations, or an application
NON_CONSTANT_ATTRIBUTES =2  # This block has non-constant attribute definitions (this bit is not set if the block has any attribute definitions that are constant, or has no attribute definitions at all)
XREF =4  # This block is an external reference (xref)
XREF_OVERLAY =8  # This block is an xref overlay
EXTERNAL =16 # This block is externally dependent
RESOLVED =32 # This is a resolved external reference, or dependent of an external reference (ignored on input)
REFERENCED =64 # This definition is a referenced external reference (ignored on input)

#---mtext flags
#attachment point
TOP_LEFT = 1
TOP_CENTER = 2
TOP_RIGHT = 3
MIDDLE_LEFT = 4
MIDDLE_CENTER = 5
MIDDLE_RIGHT    = 6
BOTTOM_LEFT = 7
BOTTOM_CENTER = 8
BOTTOM_RIGHT = 9
#drawing direction
LEFT_RIGHT = 1
TOP_BOTTOM = 3
BY_STYLE = 5 #the flow direction is inherited from the associated text style
#line spacing style (optional):
AT_LEAST = 1 #taller characters will override
EXACT = 2 #taller characters will not override

#---polyline flag 70
CLOSED =1     # This is a closed polyline (or a polygon mesh closed in the M direction)
CURVE_FIT =2      # Curve-fit vertices have been added
SPLINE_FIT =4     # Spline-fit vertices have been added
POLYLINE_3D =8    # This is a 3D polyline
POLYGON_MESH =16     # This is a 3D polygon mesh
CLOSED_N =32     # The polygon mesh is closed in the N direction
POLYFACE_MESH =64    # The polyline is a polyface mesh
CONTINOUS_LINETYPE_PATTERN =128 # The linetype pattern is generated continuously around the vertices of this polyline

#---polyline flag 75, = curve type
QUADRIC_NURBS = 5
CUBIC_NURBS = 6
BEZIER_CURVE = 8


#---text flags
#horizontal
LEFT = 0
CENTER = 1
RIGHT = 2
ALIGNED = 3 #if vertical alignment = 0
MIDDLE = 4 #if vertical alignment = 0
FIT = 5 #if vertical alignment = 0
#vertical
BASELINE = 0
BOTTOM  = 1
MIDDLE = 2
TOP = 3

####3) Classes
#---entitities -----------------------------------------------
#--------------------------
class Arc(_Entity):
    """Arc, angles in degrees."""
    def __init__(self,center=(0,0,0),radius=1,
                 startAngle=0.0,endAngle=90,**common):
        """Angles in degrees."""
        _Entity.__init__(self,**common)
        self.center=center
        self.radius=radius
        self.startAngle=startAngle
        self.endAngle=endAngle
    def __str__(self):
        return '  0\nARC\n%s%s\n 40\n%s\n 50\n%s\n 51\n%s\n'%\
               (self._common(),_point(self.center),
                self.radius,self.startAngle,self.endAngle)

#-----------------------------------------------
class Circle(_Entity):
    """Circle"""
    def __init__(self,center=(0,0,0),radius=1,**common):
        _Entity.__init__(self,**common)
        self.center=center
        self.radius=radius
    def __str__(self):
        return '  0\nCIRCLE\n%s%s\n 40\n%s\n'%\
               (self._common(),_point(self.center),self.radius)

#-----------------------------------------------
class Face(_Entity):
    """3dface"""
    def __init__(self,points,**common):
        _Entity.__init__(self,**common)
        while len(points)<4: #fix for r12 format
            points.append(points[-1])
        self.points=points

    def __str__(self):
        out = '  0\n3DFACE\n%s%s\n' %(self._common(),_points(self.points))
        return out

#-----------------------------------------------
class Insert(_Entity):
    """Block instance."""
    def __init__(self,name,point=(0,0,0),
                 xscale=None,yscale=None,zscale=None,
                 cols=None,colspacing=None,rows=None,rowspacing=None,
                 rotation=None,
                 **common):
        _Entity.__init__(self,**common)
        self.name=name
        self.point=point
        self.xscale=xscale
        self.yscale=yscale
        self.zscale=zscale
        self.cols=cols
        self.colspacing=colspacing
        self.rows=rows
        self.rowspacing=rowspacing
        self.rotation=rotation

    def __str__(self):
        result='  0\nINSERT\n  2\n%s\n%s%s\n'%\
                (self.name,self._common(),_point(self.point))
        if self.xscale!=None:result+=' 41\n%s\n'%self.xscale
        if self.yscale!=None:result+=' 42\n%s\n'%self.yscale
        if self.zscale!=None:result+=' 43\n%s\n'%self.zscale
        if self.rotation:result+=' 50\n%s\n'%self.rotation
        if self.cols!=None:result+=' 70\n%s\n'%self.cols
        if self.colspacing!=None:result+=' 44\n%s\n'%self.colspacing
        if self.rows!=None:result+=' 71\n%s\n'%self.rows
        if self.rowspacing!=None:result+=' 45\n%s\n'%self.rowspacing
        return result

#-----------------------------------------------
class Line(_Entity):
    """Line"""
    def __init__(self,points,**common):
        _Entity.__init__(self,**common)
        self.points=points
    def __str__(self):
        return '  0\nLINE\n%s%s\n' %(
                self._common(), _points(self.points))


#-----------------------------------------------
class PolyLine(_Entity):
    def __init__(self,points,org_point=[0,0,0],flag70=0,flag75=0,width=None,**common):
        #width = number, or width = list [width_start=None, width_end=None]
        #for 2d-polyline: points = [ [[x, y, z], vflag=None, [width_start=None, width_end=None], bulge=0 or None] ...]
        #for 3d-polyline: points = [ [[x, y, z], vflag=None], ...]
        #for polyface: points = [points_list, faces_list]
        _Entity.__init__(self,**common)
        self.points=points
        self.org_point=org_point
        self.pflag70 = flag70
        self.pflag75 = flag75
        self.polyface = False
        self.polyline2d = False
        self.faces = [] # dummy value
        self.width= None # dummy value
        if self.pflag70 & POLYFACE_MESH:
            self.polyface=True
            self.points=points[0]
            self.faces=points[1]
            self.p_count=len(self.points)
            self.f_count=len(self.faces)
        elif not (self.pflag70 & POLYLINE_3D):
            self.polyline2d = True
            if width:
                if type(width)!='list':  width=[width,width]
                self.width=width

    def __str__(self):
        result= '  0\nPOLYLINE\n%s 70\n%s\n' %(self._common(),self.pflag70)
        result+=' 66\n1\n'
        result+='%s\n' %_point(self.org_point)
        if self.polyface:
            result+=' 71\n%s\n' %self.p_count
            result+=' 72\n%s\n' %self.f_count
        elif self.polyline2d:
            if self.width!=None: result+=' 40\n%s\n 41\n%s\n' %(self.width[0],self.width[1])
        if self.pflag75:
            result+=' 75\n%s\n' %self.pflag75
        for point in self.points:
            result+='  0\nVERTEX\n'
            result+='  8\n%s\n' %self.layer
            if self.polyface:
                result+='%s\n' %_point(point)
                result+=' 70\n192\n'
            elif self.polyline2d:
                result+='%s\n' %_point(point[0])
                flag = point[1]
                if len(point)>2:
                    [width1, width2] = point[2]
                    if width1!=None: result+=' 40\n%s\n' %width1
                    if width2!=None: result+=' 41\n%s\n' %width2
                if len(point)==4:
                    bulge = point[3]
                    if bulge: result+=' 42\n%s\n' %bulge
                if flag:
                        result+=' 70\n%s\n' %flag
            else:
                result+='%s\n' %_point(point[0])
                flag = point[1]
                if flag:
                        result+=' 70\n%s\n' %flag

        for face in self.faces:
            result+='  0\nVERTEX\n'
            result+='  8\n%s\n' %self.layer
            result+='%s\n' %_point(self.org_point)
            result+=' 70\n128\n'
            result+=' 71\n%s\n' %face[0]
            result+=' 72\n%s\n' %face[1]
            result+=' 73\n%s\n' %face[2]
            if len(face)==4: result+=' 74\n%s\n' %face[3]
        result+='  0\nSEQEND\n'
        result+='  8\n%s\n' %self.layer
        return result

#-----------------------------------------------
class Point(_Entity):
    """Point."""
    def __init__(self,points=None,**common):
        _Entity.__init__(self,**common)
        self.points=points
    def __str__(self): # TODO:
        return '  0\nPOINT\n%s%s\n' %(self._common(),
             _points(self.points)
            )

#-----------------------------------------------
class Solid(_Entity):
    """Colored solid fill."""
    def __init__(self,points=None,**common):
        _Entity.__init__(self,**common)
        self.points=points
    def __str__(self):
        return '  0\nSOLID\n%s%s\n' %(self._common(),
             _points(self.points[:2]+[self.points[3],self.points[2]])
            )


#-----------------------------------------------
class Text(_Entity):
    """Single text line."""
    def __init__(self,text='',point=(0,0,0),alignment=None,
                 flag=None,height=1,justifyhor=None,justifyver=None,
                 rotation=None,obliqueAngle=None,style=None,xscale=None,**common):
        _Entity.__init__(self,**common)
        self.text=text
        self.point=point
        self.alignment=alignment
        self.flag=flag
        self.height=height
        self.justifyhor=justifyhor
        self.justifyver=justifyver
        self.rotation=rotation
        self.obliqueAngle=obliqueAngle
        self.style=style
        self.xscale=xscale
    def __str__(self):
        result= '  0\nTEXT\n%s%s\n 40\n%s\n  1\n%s\n'%\
                (self._common(),_point(self.point),self.height,self.text)
        if self.rotation: result+=' 50\n%s\n'%self.rotation
        if self.xscale: result+=' 41\n%s\n'%self.xscale
        if self.obliqueAngle: result+=' 51\n%s\n'%self.obliqueAngle
        if self.style: result+='  7\n%s\n'%self.style
        if self.flag: result+=' 71\n%s\n'%self.flag
        if self.justifyhor: result+=' 72\n%s\n'%self.justifyhor
        if self.alignment: result+='%s\n'%_point(self.alignment,1)
        if self.justifyver: result+=' 73\n%s\n'%self.justifyver
        return result

#-----------------------------------------------
class Mtext(Text):
    """Surrogate for mtext, generates some Text instances."""
    def __init__(self,text='',point=(0,0,0),width=250,spacingFactor=1.5,down=0,spacingWidth=None,**options):
        Text.__init__(self,text=text,point=point,**options)
        if down:spacingFactor*=-1
        self.spacingFactor=spacingFactor
        self.spacingWidth=spacingWidth
        self.width=width
        self.down=down
    def __str__(self):
        texts=self.text.replace('\r\n','\n').split('\n')
        if not self.down:texts.reverse()
        result=''
        x=y=0
        if self.spacingWidth:spacingWidth=self.spacingWidth
        else:spacingWidth=self.height*self.spacingFactor
        for text in texts:
            while text:
                result+='%s' %Text(text[:self.width],
                    point=(self.point[0]+x*spacingWidth,
                           self.point[1]+y*spacingWidth,
                           self.point[2]),
                    alignment=self.alignment,flag=self.flag,height=self.height,
                    justifyhor=self.justifyhor,justifyver=self.justifyver,
                    rotation=self.rotation,obliqueAngle=self.obliqueAngle,
                    style=self.style,xscale=self.xscale,parent=self
                )
                text=text[self.width:]
                if self.rotation:x+=1
                else:y+=1
        return result[1:]

#-----------------------------------------------
##class _Mtext(_Entity):
    """Mtext not functioning for minimal dxf."""
    """
    def __init__(self,text='',point=(0,0,0),attachment=1,
                 charWidth=None,charHeight=1,direction=1,height=100,rotation=0,
                 spacingStyle=None,spacingFactor=None,style=None,width=100,
                 xdirection=None,**common):
        _Entity.__init__(self,**common)
        self.text=text
        self.point=point
        self.attachment=attachment
        self.charWidth=charWidth
        self.charHeight=charHeight
        self.direction=direction
        self.height=height
        self.rotation=rotation
        self.spacingStyle=spacingStyle
        self.spacingFactor=spacingFactor
        self.style=style
        self.width=width
        self.xdirection=xdirection
    def __str__(self):
        input=self.text
        text=''
        while len(input)>250:
            text+='3\n%s\n'%input[:250]
            input=input[250:]
        text+='1\n%s\n'%input
        result= '0\nMTEXT\n%s\n%s\n40\n%s\n41\n%s\n71\n%s\n72\n%s%s\n43\n%s\n50\n%s\n'%\
                (self._common(),_point(self.point),self.charHeight,self.width,
                 self.attachment,self.direction,text,
                 self.height,
                 self.rotation)
        if self.style:result+='7\n%s\n'%self.style
        if self.xdirection:result+='%s\n'%_point(self.xdirection,1)
        if self.charWidth:result+='42\n%s\n'%self.charWidth
        if self.spacingStyle:result+='73\n%s\n'%self.spacingStyle
        if self.spacingFactor:result+='44\n%s\n'%self.spacingFactor
        return result
    """


#---tables ---------------------------------------------------
#-----------------------------------------------
class Block(_Collection):
    """Use list methods to add entities, eg append."""
    def __init__(self,name,layer='0',flag=0,base=(0,0,0),entities=[]):
        self.entities=copy.copy(entities)
        _Collection.__init__(self,entities)
        self.layer=layer
        self.name=name
        self.flag=0
        self.base=base
    def __str__(self): # TODO:
        e=''.join([str(x)for x in self.entities])
        return '  0\nBLOCK\n  8\n%s\n  2\n%s\n 70\n%s\n%s\n  3\n%s\n%s  0\nENDBLK\n'%\
               (self.layer,self.name.upper(),self.flag,_point(self.base),self.name.upper(),e)

#-----------------------------------------------
class Layer(_Call):
    """Layer"""
    def __init__(self,name='pydxf',color=7,lineType='continuous',flag=64):
        self.name=name
        self.color=color
        self.lineType=lineType
        self.flag=flag
    def __str__(self):
        return '  0\nLAYER\n  2\n%s\n 70\n%s\n 62\n%s\n  6\n%s\n'%\
               (self.name.upper(),self.flag,self.color,self.lineType)

#-----------------------------------------------
class LineType(_Call):
    """Custom linetype"""
    def __init__(self,name='CONTINUOUS',description='Solid line',elements=[0.0],flag=0):
        self.name=name
        self.description=description
        self.elements=copy.copy(elements)
        self.flag=flag
    def __str__(self):
        result = '  0\nLTYPE\n  2\n%s\n 70\n%s\n  3\n%s\n 72\n65\n'%\
            (self.name.upper(),self.flag,self.description)
        if self.elements:
            elements = ' 73\n%s\n' %(len(self.elements)-1)
            elements += ' 40\n%s\n' %(self.elements[0])
            for e in self.elements[1:]:
                elements += ' 49\n%s\n' %e
            result += elements
        return result


#-----------------------------------------------
class Style(_Call):
    """Text style"""
    def __init__(self,name='standard',flag=0,height=0,widthFactor=1.0,obliqueAngle=0.0,
                 mirror=0,lastHeight=1,font='arial.ttf',bigFont=''):
        self.name=name
        self.flag=flag
        self.height=height
        self.widthFactor=widthFactor
        self.obliqueAngle=obliqueAngle
        self.mirror=mirror
        self.lastHeight=lastHeight
        self.font=font
        self.bigFont=bigFont
    def __str__(self):
        return '  0\nSTYLE\n  2\n%s\n 70\n%s\n 40\n%s\n 41\n%s\n 50\n%s\n 71\n%s\n 42\n%s\n 3\n%s\n 4\n%s\n'%\
               (self.name.upper(),self.flag,self.flag,self.widthFactor,
                self.obliqueAngle,self.mirror,self.lastHeight,
                self.font.upper(),self.bigFont.upper())

#-----------------------------------------------
class VPort(_Call):
    def __init__(self,name,flag=0,
                leftBottom=(0.0,0.0),
                rightTop=(1.0,1.0),
                center=(0.5,0.5),
                snap_base=(0.0,0.0),
                snap_spacing=(0.1,0.1),
                grid_spacing=(0.1,0.1),
                direction=(0.0,0.0,1.0),
                target=(0.0,0.0,0.0),
                height=1.0,
                ratio=1.0,
                lens=50.0,
                frontClipping=0.0,
                backClipping=0.0,
                snap_rotation=0.0,
                twist=0.0,
                mode=0,
                circle_zoom=100,
                fast_zoom=1,
                ucsicon=1,
                snap_on=0,
                grid_on=0,
                snap_style=0,
                snap_isopair=0
                ):
        self.name=name
        self.flag=flag
        self.leftBottom=leftBottom
        self.rightTop=rightTop
        self.center=center
        self.snap_base=snap_base
        self.snap_spacing=snap_spacing
        self.grid_spacing=grid_spacing
        self.direction=direction
        self.target=target
        self.height=float(height)
        self.ratio=float(ratio)
        self.lens=float(lens)
        self.frontClipping=float(frontClipping)
        self.backClipping=float(backClipping)
        self.snap_rotation=float(snap_rotation)
        self.twist=float(twist)
        self.mode=mode
        self.circle_zoom=circle_zoom
        self.fast_zoom=fast_zoom
        self.ucsicon=ucsicon
        self.snap_on=snap_on
        self.grid_on=grid_on
        self.snap_style=snap_style
        self.snap_isopair=snap_isopair
    def __str__(self):
        output = ['  0', 'VPORT',
            '  2', self.name,
            ' 70', self.flag,
            _point(self.leftBottom),
            _point(self.rightTop,1),
            _point(self.center,2), # View center point (in DCS)
            _point(self.snap_base,3),
            _point(self.snap_spacing,4),
            _point(self.grid_spacing,5),
            _point(self.direction,6), #view direction from target (in WCS)
            _point(self.target,7),
            ' 40', self.height,
            ' 41', self.ratio,
            ' 42', self.lens,
            ' 43', self.frontClipping,
            ' 44', self.backClipping,
            ' 50', self.snap_rotation,
            ' 51', self.twist,
            ' 71', self.mode,
            ' 72', self.circle_zoom,
            ' 73', self.fast_zoom,
            ' 74', self.ucsicon,
            ' 75', self.snap_on,
            ' 76', self.grid_on,
            ' 77', self.snap_style,
            ' 78', self.snap_isopair
            ]

        output_str = ''
        for s in output:
            output_str += '%s\n' %s
        return output_str



#-----------------------------------------------
class View(_Call):
    def __init__(self,name,flag=0,
            width=1.0,
            height=1.0,
            center=(0.5,0.5),
            direction=(0.0,0.0,1.0),
            target=(0.0,0.0,0.0),
            lens=50.0,
            frontClipping=0.0,
            backClipping=0.0,
            twist=0.0,mode=0
            ):
        self.name=name
        self.flag=flag
        self.width=float(width)
        self.height=float(height)
        self.center=center
        self.direction=direction
        self.target=target
        self.lens=float(lens)
        self.frontClipping=float(frontClipping)
        self.backClipping=float(backClipping)
        self.twist=float(twist)
        self.mode=mode
    def __str__(self):
        output = ['  0', 'VIEW',
            '  2', self.name,
            ' 70', self.flag,
            ' 40', self.height,
            _point(self.center),
            ' 41', self.width,
            _point(self.direction,1),
            _point(self.target,2),
            ' 42', self.lens,
            ' 43', self.frontClipping,
            ' 44', self.backClipping,
            ' 50', self.twist,
            ' 71', self.mode
            ]
        output_str = ''
        for s in output:
            output_str += '%s\n' %s
        return output_str

#-----------------------------------------------
def ViewByWindow(name,leftBottom=(0,0),rightTop=(1,1),**options):
    width=abs(rightTop[0]-leftBottom[0])
    height=abs(rightTop[1]-leftBottom[1])
    center=((rightTop[0]+leftBottom[0])*0.5,(rightTop[1]+leftBottom[1])*0.5)
    return View(name=name,width=width,height=height,center=center,**options)

#---drawing
#-----------------------------------------------
class Drawing(_Collection):
    """Dxf drawing. Use append or any other list methods to add objects."""
    def __init__(self,insbase=(0.0,0.0,0.0),extmin=(0.0,0.0,0.0),extmax=(0.0,0.0,0.0),
                 layers=[Layer()],linetypes=[LineType()],styles=[Style()],blocks=[],
                 views=[],vports=[],entities=None,fileName='test.dxf'):
        # TODO: replace list with None,arial
        if not entities:
            entities=[]
        _Collection.__init__(self,entities)
        self.insbase=insbase
        self.extmin=extmin
        self.extmax=extmax
        self.layers=copy.copy(layers)
        self.linetypes=copy.copy(linetypes)
        self.styles=copy.copy(styles)
        self.views=copy.copy(views)
        self.vports=copy.copy(vports)
        self.blocks=copy.copy(blocks)
        self.fileName=fileName
        #print 'deb: blocks=',blocks #----------
        #private
        #self.acadver='9\n$ACADVER\n1\nAC1006\n'
        self.acadver='  9\n$ACADVER\n  1\nAC1009\n'
        """DXF AutoCAD-Release format codes:
        AC1021  2008, 2007
        AC1018  2006, 2005, 2004
        AC1015  2002, 2000i, 2000
        AC1014  R14,14.01
        AC1012  R13
        AC1009  R12,11
        AC1006  R10
        AC1004  R9
        AC1002  R2.6
        AC1.50  R2.05
        """

    def _name(self,x):
        """Helper function for self._point"""
        return '  9\n$%s\n' %x.upper()

    def _point(self,name,x):
        """Point setting from drawing like extmin,extmax,..."""
        return '%s%s' %(self._name(name),_point(x))

    def _section(self,name,x):
        """Sections like tables,blocks,entities,..."""
        if x: xstr=''.join(x)
        else: xstr=''
        return '  0\nSECTION\n  2\n%s\n%s  0\nENDSEC\n'%(name.upper(),xstr)

    def _table(self,name,x):
        """Tables like ltype,layer,style,..."""
        if x: xstr=''.join(x)
        else: xstr=''
        return '  0\nTABLE\n  2\n%s\n 70\n%s\n%s  0\nENDTAB\n'%(name.upper(),len(x),xstr)

    def __str__(self):
        """Returns drawing as dxf string."""
        header=[self.acadver]+[self._point(attr,getattr(self,attr))+'\n' for attr in _HEADER_POINTS]
        header=self._section('header',header)

        tables=[self._table('vport',[str(x) for x in self.vports]),
                self._table('ltype',[str(x) for x in self.linetypes]),
                self._table('layer',[str(x) for x in self.layers]),
                self._table('style',[str(x) for x in self.styles]),
                self._table('view',[str(x) for x in self.views]),
        ]
        tables=self._section('tables',tables)
        blocks=self._section('blocks',[str(x) for x in self.blocks])
        entities=self._section('entities',[str(x) for x in self.entities])
        all=''.join([header,tables,blocks,entities,'  0\nEOF\n'])
        return all

    def _write_section(self,file,name,data):
        file.write('  0\nSECTION\n  2\n%s\n'%name.upper())
        for x in data:
            file.write(str(x))
        file.write('  0\nENDSEC\n')

    def saveas(self,fileName,buffer=0):
        """Writes DXF file. Needs target file name. If optional parameter buffer>0, then switch to old behavior: store entire output string in RAM.
        """
        self.fileName=fileName
        if buffer: self.save()
        else: self.export()

    def save(self):
        outfile=open(self.fileName,'w')
        outfile.write(str(self))
        outfile.close()

    def export(self):
        outfile=open(self.fileName,'w')
        header=[self.acadver]+[self._point(attr,getattr(self,attr))+'\n' for attr in _HEADER_POINTS]
        self._write_section(outfile,'header',header)
        tables=[self._table('vport',[str(x) for x in self.vports]),
            self._table('ltype',[str(x) for x in self.linetypes]),
            self._table('layer',[str(x) for x in self.layers]),
            self._table('style',[str(x) for x in self.styles]),
            self._table('view',[str(x) for x in self.views]),
            ]
        self._write_section(outfile,'tables',tables)
        self._write_section(outfile,'blocks',self.blocks)
        self._write_section(outfile,'entities',self.entities)
        outfile.write('  0\nEOF\n')
        outfile.close()


#---extras
#-----------------------------------------------
class Rectangle(_Entity):
    """Rectangle, creates lines."""
    def __init__(self,point=(0,0,0),width=1,height=1,solid=None,line=1,**common):
        _Entity.__init__(self,**common)
        self.point=point
        self.width=width
        self.height=height
        self.solid=solid
        self.line=line
    def __str__(self):
        result=''
        points=[self.point,(self.point[0]+self.width,self.point[1],self.point[2]),
            (self.point[0]+self.width,self.point[1]+self.height,self.point[2]),
            (self.point[0],self.point[1]+self.height,self.point[2]),self.point]
        if self.solid:
            result+= Solid(points=points[:-1],parent=self.solid)
        if self.line:
            for i in range(4):
                result+= Line(points=[points[i],points[i+1]],parent=self)
        return result[1:]

#-----------------------------------------------
class LineList(_Entity):
    """Like polyline, but built of individual lines."""
    def __init__(self,points=[],org_point=[0,0,0],closed=0,**common):
        _Entity.__init__(self,**common)
        self.closed=closed
        self.points=copy.copy(points)
    def __str__(self):
        if self.closed:
            points=self.points+[self.points[0]]
        else: points=self.points
        result=''
        for i in range(len(points)-1):
            result+= Line(points=[points[i][0],points[i+1][0]],parent=self)
        return result[1:]

#-----------------------------------------------------
def test():
    #Blocks
    b=Block('test')
    b.append(Solid(points=[(0,0,0),(1,0,0),(1,1,0),(0,1,0)],color=1))
    b.append(Arc(center=(1,0,0),color=2))

    #Drawing
    d=Drawing()
    #tables
    d.blocks.append(b)  #table blocks
    d.styles.append(Style())  #table styles
    d.views.append(View('Normal'))  #table view
    d.views.append(ViewByWindow('Window',leftBottom=(1,0),rightTop=(2,1)))  #idem

    #entities
    d.append(Circle(center=(1,1,0),color=3))
    d.append(Face(points=[(0,0,0),(1,0,0),(1,1,0),(0,1,0)],color=4))
    d.append(Insert('test',point=(3,3,3),cols=5,colspacing=2))
    d.append(Line(points=[(0,0,0),(1,1,1)]))
    d.append(Mtext('Click on Ads\nmultiple lines with mtext',point=(1,1,1),color=5,rotation=90))
    d.append(Text('Please donate!',point=(3,0,1)))
    #d.append(Rectangle(point=(2,2,2),width=4,height=3,color=6,solid=Solid(color=2)))
    d.append(Solid(points=[(4,4,0),(5,4,0),(7,8,0),(9,9,0)],color=3))
    #d.append(PolyLine(points=[(1,1,1),(2,1,1),(2,2,1),(1,2,1)],flag=1,color=1))

    #d.saveas('c:\\test.dxf')
    d.saveas('test.dxf')

#-----------------------------------------------------
if __name__=='__main__':
    if not copy:
        Draw.PupMenu('Error%t|This script requires a full python install')
    else: test()



# SPDX-FileCopyrightText: 2022-2023 Blender Foundation
#
# SPDX-License-Identifier: GPL-2.0-or-later

# ##### BEGIN AUTOGENERATED I18N SECTION #####
# NOTE: You can safely move around this auto-generated block (with the begin/end markers!),
#       and edit the translations by hand.
#       Just carefully respect the format of the tuple!

# Tuple of tuples:
# ((msgctxt, msgid), (sources, gen_comments), (lang, translation, (is_fuzzy, comments)), ...)
translations_tuple = (
    (("*", ""),
     ((), ()),
     ("fr_FR", "Project-Id-Version: AnimAll 0.10.0 (0)\n",
               (False,
                ("Blender's translation file (po format).",
                 "Copyright (C) 2022-2023 The Blender Foundation.",
                 "This file is distributed under the same license as the Blender package.",
                 "Damien Picard <dam.pic@free.fr>, 2022."))),
    ),
    (("*", "Tab Category"),
     (("bpy.types.AnimallAddonPreferences.category",),
      ()),
     ("fr_FR", "Catégorie d’onglet",
               (False, ())),
    ),
    (("*", "Choose a name for the category of the panel"),
     (("bpy.types.AnimallAddonPreferences.category",),
      ()),
     ("fr_FR", "Choisir un nom pour la catégorie du panneau",
               (False, ())),
    ),
    (("Operator", "Clear Animation"),
     (("bpy.types.ANIM_OT_clear_animation_animall",),
      ()),
     ("fr_FR", "Effacer l’animation",
               (False, ())),
    ),
    (("*", "Delete all keyframes for this object\nIf in a specific case it doesn't work\ntry to delete the keys manually"),
     (("bpy.types.ANIM_OT_clear_animation_animall",),
      ()),
     ("fr_FR", "Supprimer toutes les images clés pour cet objet.\n"
               "En cas d’échec, essayez de les supprimer manuellement",
               (False, ())),
    ),
    (("Operator", "Insert Key"),
     (("bpy.types.ANIM_OT_insert_keyframe_animall",),
      ()),
     ("fr_FR", "Insérer une clé",
               (False, ())),
    ),
    (("*", "Insert a Keyframe"),
     (("bpy.types.ANIM_OT_insert_keyframe_animall",),
      ()),
     ("fr_FR", "Insérer une image clé",
               (False, ())),
    ),
    (("Operator", "Delete Key"),
     (("bpy.types.ANIM_OT_delete_keyframe_animall",),
      ()),
     ("fr_FR", "Supprimer l’image clé",
               (False, ())),
    ),
    (("*", "Delete a Keyframe"),
     (("bpy.types.ANIM_OT_delete_keyframe_animall",),
      ()),
     ("fr_FR", "Supprimer une image clé",
               (False, ())),
    ),
    (("Operator", "Update Attribute Animation"),
     (("bpy.types.ANIM_OT_update_attribute_animation_animall",),
      ()),
     ("fr_FR", "Mettre à jour l’animation des attributs",
               (False, ())),
    ),
    (("*", "Update attributes from the old format"),
     (("bpy.types.ANIM_OT_update_attribute_animation_animall",),
      ()),
     ("fr_FR", "Mettre à jour les attributs depuis l’ancien format",
               (False, ())),
    ),
    (("*", "Animate"),
     (("bpy.types.VIEW3D_PT_animall",),
      ()),
     ("fr_FR", "Animer",
               (False, ())),
    ),
    (("*", "Insert keyframes on active attribute values"),
     (("bpy.types.AnimallProperties.key_active_attribute",),
      ()),
     ("fr_FR", "Insérer des clés sur l’attribut actif",
               (False, ())),
    ),
    (("*", "Insert keyframes on edge bevel weight"),
     (("bpy.types.AnimallProperties.key_edge_bevel",),
      ()),
     ("fr_FR", "Insérer des clés sur les poids de biseau d’arête",
               (False, ())),
    ),
    (("*", "Edge Crease"),
     (("bpy.types.AnimallProperties.key_edge_crease",),
      ()),
     ("fr_FR", "Plis d’arêtes",
               (False, ())),
    ),
    (("*", "Insert keyframes on edge creases"),
     (("bpy.types.AnimallProperties.key_edge_crease",),
      ()),
     ("fr_FR", "Insérer des clés sur les plis d’arête",
               (False, ())),
    ),
    (("*", "Insert keyframes on face material indices"),
     (("bpy.types.AnimallProperties.key_material_index",),
      ()),
     ("fr_FR", "Insérer des clés sur les indices de matériaux",
               (False, ())),
    ),
    (("*", "Insert keyframes on point locations"),
     (("bpy.types.AnimallProperties.key_point_location",),
      ()),
     ("fr_FR", "Insérer des clés sur les positions des points",
               (False, ())),
    ),
    (("*", "Insert keyframes on point radius (Shrink/Fatten)"),
     (("bpy.types.AnimallProperties.key_radius",),
      ()),
     ("fr_FR", "Insérer des clés sur le rayon de rayon de point (épaisseur de la courbe)",
               (False, ())),
    ),
    (("*", "Key Selected Only"),
     (("bpy.types.AnimallProperties.key_selected",),
      ()),
     ("fr_FR", "Sélection uniquement",
               (False, ())),
    ),
    (("*", "Insert keyframes only on selected elements"),
     (("bpy.types.AnimallProperties.key_selected",),
      ()),
     ("fr_FR", "Insérer des images clés seulement sur les éléments sélectionnés",
               (False, ())),
    ),
    (("*", "Insert keyframes on active Shape Key layer"),
     (("bpy.types.AnimallProperties.key_shape_key",),
      ()),
     ("fr_FR", "Insérer des clés sur le calque de clé de forme actif",
               (False, ())),
    ),
    (("*", "Insert keyframes on point tilt"),
     (("bpy.types.AnimallProperties.key_tilt",),
      ()),
     ("fr_FR", "Insérer des clés sur l’inclinaison des points",
               (False, ())),
    ),
    (("*", "Insert keyframes on active UV coordinates"),
     (("bpy.types.AnimallProperties.key_uvs",),
      ()),
     ("fr_FR", "Insérer des clés sur les coordonnées UV actives",
               (False, ())),
    ),
    (("*", "Insert keyframes on vertex bevel weight"),
     (("bpy.types.AnimallProperties.key_vertex_bevel",),
      ()),
     ("fr_FR", "Insérer des clés sur les poids de biseau des sommets",
               (False, ())),
    ),
    (("*", "Insert keyframes on vertex crease weight"),
     (("bpy.types.AnimallProperties.key_vertex_crease",),
      ()),
     ("fr_FR", "Insérer des clés sur les plis de sommets",
               (False, ())),
    ),
    (("*", "Insert keyframes on active vertex group values"),
     (("bpy.types.AnimallProperties.key_vertex_group",),
      ()),
     ("fr_FR", "Insérer des clés sur les valeurs des groupes de sommets",
               (False, ())),
    ),
    (("*", "AnimAll"),
     (("Add-on AnimAll info: name",),
      ()),
     ("fr_FR", "AnimAll",
               (False, ())),
    ),
    (("*", "Key:"),
     (("scripts/addons/animation_animall/__init__.py:200",),
      ()),
     ("fr_FR", "Insérer :",
               (False, ())),
    ),
    (("*", "Tab Category:"),
     (("scripts/addons/animation_animall/__init__.py:704",),
      ()),
     ("fr_FR", "Catégorie d’onglet :",
               (False, ())),
    ),
    (("*", "Points"),
     (("scripts/addons/animation_animall/__init__.py:206",
       "scripts/addons/animation_animall/__init__.py:213",
       "scripts/addons/animation_animall/__init__.py:243"),
      ()),
     ("fr_FR", "Points",
               (False, ())),
    ),
    (("*", "Others"),
     (("scripts/addons/animation_animall/__init__.py:209",
       "scripts/addons/animation_animall/__init__.py:226",
       "scripts/addons/animation_animall/__init__.py:251"),
      ()),
     ("fr_FR", "Autres",
               (False, ())),
    ),
    (("*", "Bevel"),
     (("scripts/addons/animation_animall/__init__.py:215",
       "scripts/addons/animation_animall/__init__.py:220"),
      ()),
     ("fr_FR", "Biseau",
               (False, ())),
    ),
    (("*", "Edges"),
     (("scripts/addons/animation_animall/__init__.py:219",),
      ()),
     ("fr_FR", "Arêtes",
               (False, ())),
    ),
    (("*", "Crease"),
     (("scripts/addons/animation_animall/__init__.py:216",
       "scripts/addons/animation_animall/__init__.py:221",),
      ()),
     ("fr_FR", "Plis",
               (False, ())),
    ),
    (("*", "Faces"),
     (("scripts/addons/animation_animall/__init__.py:223",),
      ()),
     ("fr_FR", "Faces",
               (False, ())),
    ),
    (("*", "\"Location\" and \"Shape Key\" are redundant?"),
     (("scripts/addons/animation_animall/__init__.py:273",),
      ()),
     ("fr_FR", "\"Position\" et \"Clé de forme\" sont redondants ?",
               (False, ())),
    ),
    (("*", "Splines"),
     (("scripts/addons/animation_animall/__init__.py:248",),
      ()),
     ("fr_FR", "Splines",
               (False, ())),
    ),
    (("*", "Maybe set \"%s\" to 1.0?"),
     (("scripts/addons/animation_animall/__init__.py:264"),
      ()),
     ("fr_FR", "Essayez de mettre « %s » à 1.0 ?",
               (False, ())),
    ),
    (("*", "Cannot key on Basis Shape"),
     (("scripts/addons/animation_animall/__init__.py:267",),
      ()),
     ("fr_FR", "Impossible d’ajouter une clé sur la forme de base",
               (False, ())),
    ),
    (("*", "No active Shape Key"),
     (("scripts/addons/animation_animall/__init__.py:270",),
      ()),
     ("fr_FR", "Pas de clé de forme active",
               (False, ())),
    ),
    (("*", "Clear Animation could not be performed"),
     (("scripts/addons/animation_animall/__init__.py:615",),
      ()),
     ("fr_FR", "La suppression de l’animation n’a pas pu aboutir",
               (False, ())),
    ),
    (("*", "Object includes old-style attributes. Consider updating them."),
     (("scripts/addons/animation_animall/__init__.py:237",),
      ()),
     ("fr_FR", "L’objet contient des attributs à l’ancien format. Veuillez les mettre à jour.",
               (False, ())),
    ),
    (("*", "Vertex %s"),
     (("scripts/addons/animation_animall/__init__.py:141",
       "scripts/addons/animation_animall/__init__.py:368",
       "scripts/addons/animation_animall/__init__.py:382",
       "scripts/addons/animation_animall/__init__.py:416"),
      ()),
     ("fr_FR", "Sommet %s",
               (False, ())),
    ),
    (("*", "Edge %s"),
     (("scripts/addons/animation_animall/__init__.py:143"),
      ()),
     ("fr_FR", "Arête %s",
               (False, ())),
    ),
    (("*", "Point %s"),
     (("scripts/addons/animation_animall/__init__.py:320",),
      ()),
     ("fr_FR", "Point %s",
               (False, ())),
    ),
    (("*", "Spline %s"),
     (("scripts/addons/animation_animall/__init__.py:328",),
      ()),
     ("fr_FR", "Spline %s",
               (False, ())),
    ),
    (("*", "Face %s"),
     (("scripts/addons/animation_animall/__init__.py:145",
       "scripts/addons/animation_animall/__init__.py:395"),
      ()),
     ("fr_FR", "Face %s",
               (False, ())),
    ),
    (("*", "%s Point %s"),
     (("scripts/addons/animation_animall/__init__.py:315",),
      ()),
     ("fr_FR", "%s Point %s",
               (False, ())),
    ),
    (("*", "Loop %s"),
     (("scripts/addons/animation_animall/__init__.py:147",),
      ()),
     ("fr_FR", "Boucle %s",
               (False, ())),
    ),
    (("*", "UV Layer %s"),
     (("scripts/addons/animation_animall/__init__.py:409",),
      ()),
     ("fr_FR", "Calque UV %s",
               (False, ())),
    ),
    (("*", "%s Vertex %s"),
     (("scripts/addons/animation_animall/__init__.py:416",),
      ()),
     ("fr_FR", "%s Sommet %s",
               (False, ())),
    ),
    (("*", "Spline %s CV %s"),
     (("scripts/addons/animation_animall/__init__.py:338",
       "scripts/addons/animation_animall/__init__.py:339",
       "scripts/addons/animation_animall/__init__.py:340",
       "scripts/addons/animation_animall/__init__.py:343",
       "scripts/addons/animation_animall/__init__.py:346",
       "scripts/addons/animation_animall/__init__.py:352",
       "scripts/addons/animation_animall/__init__.py:355",
       "scripts/addons/animation_animall/__init__.py:358"),
      ()),
     ("fr_FR", "Spline %s Point %s",
               (False, ())),
    ),
    (("*", "%s Spline %s CV %s"),
     (("scripts/addons/animation_animall/__init__.py:432",
       "scripts/addons/animation_animall/__init__.py:434",
       "scripts/addons/animation_animall/__init__.py:437",
       "scripts/addons/animation_animall/__init__.py:440",
       "scripts/addons/animation_animall/__init__.py:442",
       "scripts/addons/animation_animall/__init__.py:450",
       "scripts/addons/animation_animall/__init__.py:452",
       "scripts/addons/animation_animall/__init__.py:453"),
      ()),
     ("fr_FR", "%s Spline %s Point %s",
               (False, ())),
    ),
    (("*", "3D View > Toolbox > Animation tab > AnimAll"),
     (("Add-on AnimAll info: location",),
      ()),
     ("fr_FR", "Vue 3D > Panneau N > Onglet Animer > AnimAll",
               (False, ())),
    ),
    (("*", "Allows animation of mesh, lattice, curve and surface data"),
     (("Add-on AnimAll info: description",),
      ()),
     ("fr_FR", "Permet d’animer les données de maillages, de lattices, de courbes et de surfaces",
               (False, ())),
    ),
)

translations_dict = {}
for msg in translations_tuple:
    key = msg[0]
    for lang, trans, (is_fuzzy, comments) in msg[2:]:
        if trans and not is_fuzzy:
            translations_dict.setdefault(lang, {})[key] = trans

# ##### END AUTOGENERATED I18N SECTION #####



# SPDX-FileCopyrightText: 2019-2022 Blender Foundation
#
# SPDX-License-Identifier: GPL-2.0-or-later

bl_info = {
    'name': 'Curve Fillet',
    'author': 'Spivak Vladimir (cwolf3d)',
    'version': (0, 0, 1),
    'blender': (2, 80, 0),
    'location': 'Curve Tools addon. (N) Panel',
    'description': 'Various types of fillet (chamfering)',
    'warning': '', # used for warning icon and text in addons panel
    'doc_url': '',
    'tracker_url': '',
    'category': 'Curve',
}


import bpy
from bpy.props import *
from bpy_extras import object_utils, view3d_utils
from mathutils import  *
from math import  *

def click(self, context, event):
    bpy.ops.object.mode_set(mode = 'EDIT')
    bpy.context.view_layer.update()


def remove_handler(handlers):
    for handler in handlers:
        try:
            bpy.types.SpaceView3D.draw_handler_remove(handler, 'WINDOW')
        except:
            pass
    for handler in handlers:
        handlers.remove(handler)


class Fillet(bpy.types.Operator):
    bl_idname = "curvetools.fillet"
    bl_label = "Curve Fillet"
    bl_description = "Curve Fillet"
    bl_options = {'REGISTER', 'UNDO'}

    x: IntProperty(name="x", description="x")
    y: IntProperty(name="y", description="y")
    location3D: FloatVectorProperty(name = "",
                description = "Start location",
                default = (0.0, 0.0, 0.0),
                subtype = 'XYZ')

    handlers = []

    def execute(self, context):
        self.report({'INFO'}, "ESC or TAB - cancel")
        bpy.ops.object.mode_set(mode = 'EDIT')

        # color change in the panel
        self.path_color = bpy.context.scene.curvetools.path_color
        self.path_thickness = bpy.context.scene.curvetools.path_thickness

    def modal(self, context, event):
        context.area.tag_redraw()

        if event.type in {'ESC', 'TAB'}:  # Cancel
            remove_handler(self.handlers)
            return {'CANCELLED'}

        if event.type in {'X', 'DEL'}:  # Cancel
            remove_handler(self.handlers)
            bpy.ops.curve.delete(type='VERT')
            return {'RUNNING_MODAL'}

        elif event.alt and event.shift and event.type == 'LEFTMOUSE':
            click(self, context, event)

        elif event.alt and not event.shift and event.type == 'LEFTMOUSE':
            remove_handler(self.handlers)
            bpy.ops.curve.select_all(action='DESELECT')
            click(self, context, event)

        elif event.alt and event.type == 'RIGHTMOUSE':
           remove_handler(self.handlers)
           bpy.ops.curve.select_all(action='DESELECT')
           click(self, context, event)

        elif event.alt and not event.shift and event.shift and event.type == 'RIGHTMOUSE':
            click(self, context, event)

        elif event.type == 'A':
            remove_handler(self.handlers)
            bpy.ops.curve.select_all(action='DESELECT')

        elif event.type == 'MOUSEMOVE':  #
            self.x = event.mouse_x
            self.y = event.mouse_y
            region = bpy.context.region
            rv3d = bpy.context.space_data.region_3d
            self.location3D = view3d_utils.region_2d_to_location_3d(
                region,
                rv3d,
                (event.mouse_region_x, event.mouse_region_y),
                (0.0, 0.0, 0.0)
                )

        return {'PASS_THROUGH'}

    def invoke(self, context, event):
        self.execute(context)
        context.window_manager.modal_handler_add(self)
        return {'RUNNING_MODAL'}

    @classmethod
    def poll(cls, context):
        return (context.object is not None and
                context.object.type == 'CURVE')

def register():
    for cls in classes:
        bpy.utils.register_class(operators)

def unregister():
    for cls in classes:
        bpy.utils.unregister_class(operators)

if __name__ == "__main__":
    register()

operators = [Fillet]



# SPDX-FileCopyrightText: 2011-2022 Blender Foundation
#
# SPDX-License-Identifier: GPL-2.0-or-later

import bpy
from . import presets
from . import data as data_types


class RENDER_UL_copy_settings(bpy.types.UIList):
    def draw_item(self, context, layout, data, item, icon, active_data, active_propname, index):
        #assert(isinstance(item, (data_types.RenderCopySettingsScene, data_types.RenderCopySettingsDataSetting)))
        if self.layout_type in {'DEFAULT', 'COMPACT'}:
            if isinstance(item, data_types.RenderCopySettingsDataSetting):
                layout.label(text=item.name, icon_value=icon)
                layout.prop(item, "copy", text="")
            else: #elif isinstance(item, data_types.RenderCopySettingsDataScene):
                layout.prop(item, "allowed", text=item.name, toggle=True)
        elif self.layout_type in {'GRID'}:
            layout.alignment = 'CENTER'
            if isinstance(item, data_types.RenderCopySettingsDataSetting):
                layout.label(text=item.name, icon_value=icon)
                layout.prop(item, "copy", text="")
            else: #elif isinstance(item, data_types.RenderCopySettingsDataScene):
                layout.prop(item, "allowed", text=item.name, toggle=True)


class RENDER_PT_copy_settings(bpy.types.Panel):
    bl_label = "Copy Settings"
    bl_space_type = "PROPERTIES"
    bl_region_type = "WINDOW"
    bl_context = "render"
    bl_options = {'DEFAULT_CLOSED'}
    COMPAT_ENGINES = {'BLENDER_RENDER'}

    def draw(self, context):
        layout = self.layout
        cp_sett = context.scene.render_copy_settings

        layout.operator("scene.render_copy_settings", text="Copy Render Settings")

        split = layout.split(factor=0.75)
        split.template_list("RENDER_UL_copy_settings", "settings", cp_sett, "affected_settings",
                            cp_sett, "affected_settings_idx", rows=5)

        col = split.column()
        all_set = {sett.strid for sett in cp_sett.affected_settings if sett.copy}
        for p in presets.presets:
            label = ""
            if p.elements & all_set == p.elements:
                label = "Clear {}".format(p.ui_name)
            else:
                label = "Set {}".format(p.ui_name)
            col.operator("scene.render_copy_settings_preset", text=label).presets = {p.rna_enum[0]}

        layout.prop(cp_sett, "filter_scene")
        if len(cp_sett.allowed_scenes):
            layout.label(text="Affected Scenes:")
            layout.template_list("RENDER_UL_copy_settings", "scenes", cp_sett, "allowed_scenes",
#                                 cp_sett, "allowed_scenes_idx", rows=6, type='GRID')
                                 cp_sett, "allowed_scenes_idx", rows=6) # XXX Grid is not nice currently...
        else:
            layout.label(text="No Affectable Scenes!", icon="ERROR")


classes = (
    RENDER_UL_copy_settings,
    RENDER_PT_copy_settings,
)



# SPDX-FileCopyrightText: 2017-2023 Blender Foundation
#
# SPDX-License-Identifier: GPL-2.0-or-later

# --------------------------------- DUAL MESH -------------------------------- #
# -------------------------------- version 0.3 ------------------------------- #
#                                                                              #
# Convert a generic mesh to its dual. With open meshes it can get some wired   #
# effect on the borders.                                                       #
#                                                                              #
#                        (c)   Alessandro Zomparelli                           #
#                                    (2017)                                    #
#                                                                              #
# http://www.co-de-it.com/                                                     #
#                                                                              #
# ############################################################################ #


import bpy
from bpy.types import Operator
from bpy.props import (
        BoolProperty,
        EnumProperty,
        )
import bmesh
from .utils import *


class dual_mesh_tessellated(Operator):
    bl_idname = "object.dual_mesh_tessellated"
    bl_label = "Dual Mesh"
    bl_description = ("Generate a polygonal mesh using Tessellate. (Non-destructive)")
    bl_options = {'REGISTER', 'UNDO'}

    apply_modifiers : BoolProperty(
        name="Apply Modifiers",
        default=True,
        description="Apply object's modifiers"
        )

    source_faces : EnumProperty(
            items=[
                ('QUAD', 'Quad Faces', ''),
                ('TRI', 'Triangles', '')],
            name="Source Faces",
            description="Triangles works with any geometry." \
                        "Quad option is faster when the object has only Quads",
            default="TRI",
            options={'LIBRARY_EDITABLE'}
            )

    link_component : BoolProperty(
        name="Editable Component",
        default=False,
        description="Add Component Object to the Scene"
        )

    def execute(self, context):
        auto_layer_collection()
        ob0 = context.object
        name1 = "DualMesh_{}_Component".format(self.source_faces)
        # Generate component
        if self.source_faces == 'QUAD':
            verts = [(1.0, 0.0, 0.0), (0.5, 0.0, 0.0),
                    (0.0, 0.0, 0.0), (0.0, 0.5, 0.0),
                    (0.0, 1.0, 0.0), (0.5, 1.0, 0.0),
                    (1.0, 1.0, 0.0), (1.0, 0.5, 0.0),
                    (2/3, 1/3, 0.0), (1/3, 2/3, 0.0)]
            edges = [(0,1), (1,2), (2,3), (3,4), (4,5), (5,6), (6,7),
                        (7,0), (1,8), (8,7), (3,9), (9,5), (8,9)]
            faces = [(7,8,1,0), (8,9,3,2,1), (9,5,4,3), (9,8,7,6,5)]
        else:
            verts = [(0.0, 0.0, 0.0), (1.0, 0.0, 0.0),
                    (0.0, 1.0, 0.0), (1.0, 1.0, 0.0),
                    (0.5, 1/3, 0.0), (0.0, 0.5, 0.0),
                    (1.0, 0.5, 0.0), (0.5, 0.0, 0.0)]
            edges = [(0,5), (1,7), (3,6), (2,3), (2,5), (1,6), (0,7),
                        (4,5), (4,7), (4,6)]
            faces = [(5,0,7,4), (7,1,6,4), (3,2,5,4,6)]

        # check pre-existing component
        try:
            _verts = [0]*len(verts)*3
            __verts = [c for co in verts for c in co]
            ob1 = bpy.data.objects[name1]
            ob1.data.vertices.foreach_get("co",_verts)
            for a, b in zip(_verts, __verts):
                if abs(a-b) > 0.0001:
                    raise ValueError
        except:
            me = bpy.data.meshes.new("Dual-Mesh")  # add a new mesh
            me.from_pydata(verts, edges, faces)
            me.update(calc_edges=True, calc_edges_loose=True)
            if self.source_faces == 'QUAD': seams = (0,1,2,3,4,5,6,9)
            else: seams = (0,1,2,3,4,5,7)
            for i in seams: me.edges[i].use_seam = True
            ob1 = bpy.data.objects.new(name1, me)
            # fix visualization issue
            if self.link_component:
                context.collection.objects.link(ob1)
                context.view_layer.objects.active = ob1
                ob1.select_set(True)
                bpy.ops.object.editmode_toggle()
                bpy.ops.object.editmode_toggle()
                ob1.select_set(False)
                ob1.hide_render = True
        ob = convert_object_to_mesh(ob0,False,False)
        ob.name = 'DualMesh'
        ob.tissue.tissue_type = 'TESSELLATE'
        ob.tissue.bool_lock = True
        ob.tissue_tessellate.component = ob1
        ob.tissue_tessellate.generator = ob0
        ob.tissue_tessellate.gen_modifiers = self.apply_modifiers
        ob.tissue_tessellate.merge = True
        ob.tissue_tessellate.bool_dissolve_seams = True
        if self.source_faces == 'TRI': ob.tissue_tessellate.fill_mode = 'TRI'
        bpy.ops.object.tissue_update_tessellate()
        ob.tissue.bool_lock = False
        ob.location = ob0.location
        ob.matrix_world = ob0.matrix_world
        return {'FINISHED'}

    def invoke(self, context, event):
        return context.window_manager.invoke_props_dialog(self)

class dual_mesh(Operator):
    bl_idname = "object.dual_mesh"
    bl_label = "Convert to Dual Mesh"
    bl_description = ("Convert a generic mesh into a polygonal mesh. (Destructive)")
    bl_options = {'REGISTER', 'UNDO'}

    quad_method : EnumProperty(
            items=[('BEAUTY', 'Beauty',
                    'Split the quads in nice triangles, slower method'),
                    ('FIXED', 'Fixed',
                    'Split the quads on the 1st and 3rd vertices'),
                    ('FIXED_ALTERNATE', 'Fixed Alternate',
                    'Split the quads on the 2nd and 4th vertices'),
                    ('SHORTEST_DIAGONAL', 'Shortest Diagonal',
                    'Split the quads based on the distance between the vertices')
                    ],
            name="Quad Method",
            description="Method for splitting the quads into triangles",
            default="FIXED",
            options={'LIBRARY_EDITABLE'}
            )
    polygon_method : EnumProperty(
            items=[
                ('BEAUTY', 'Beauty', 'Arrange the new triangles evenly'),
                ('CLIP', 'Clip',
                 'Split the N-gon with an ear clipping algorithm')],
            name="N-gon Method",
            description="Method for splitting the N-gons into triangles",
            default="BEAUTY",
            options={'LIBRARY_EDITABLE'}
            )
    preserve_borders : BoolProperty(
            name="Preserve Borders",
            default=True,
            description="Preserve original borders"
            )
    apply_modifiers : BoolProperty(
            name="Apply Modifiers",
            default=True,
            description="Apply object's modifiers"
            )

    def execute(self, context):
        mode = context.mode
        if mode == 'EDIT_MESH':
            mode = 'EDIT'
        act = context.active_object
        if mode != 'OBJECT':
            sel = [act]
            bpy.ops.object.mode_set(mode='OBJECT')
        else:
            sel = context.selected_objects
        doneMeshes = []

        for ob0 in sel:
            if ob0.type != 'MESH':
                continue
            if ob0.data.name in doneMeshes:
                continue
            ob = ob0
            mesh_name = ob0.data.name

            # store linked objects
            clones = []
            n_users = ob0.data.users
            count = 0
            for o in bpy.data.objects:
                if o.type != 'MESH':
                    continue
                if o.data.name == mesh_name:
                    count += 1
                    clones.append(o)
                if count == n_users:
                    break

            if self.apply_modifiers:
                bpy.ops.object.convert(target='MESH')
            ob.data = ob.data.copy()
            bpy.ops.object.select_all(action='DESELECT')
            ob.select_set(True)
            context.view_layer.objects.active = ob0
            bpy.ops.object.mode_set(mode='EDIT')

            # prevent borders erosion
            bpy.ops.mesh.select_mode(
                    use_extend=False, use_expand=False, type='EDGE'
                    )
            bpy.ops.mesh.select_non_manifold(
                    extend=False, use_wire=False, use_boundary=True,
                    use_multi_face=False, use_non_contiguous=False,
                    use_verts=False
                    )
            bpy.ops.mesh.extrude_region_move(
                    MESH_OT_extrude_region={"mirror": False},
                    TRANSFORM_OT_translate={"value": (0, 0, 0)}
                    )

            bpy.ops.mesh.select_mode(
                    use_extend=False, use_expand=False, type='VERT',
                    action='TOGGLE'
                    )
            bpy.ops.mesh.select_all(action='SELECT')
            bpy.ops.mesh.quads_convert_to_tris(
                    quad_method=self.quad_method, ngon_method=self.polygon_method
                    )
            bpy.ops.mesh.select_all(action='DESELECT')
            bpy.ops.object.mode_set(mode='OBJECT')
            subsurf_modifier = context.object.modifiers.new("dual_mesh_subsurf", 'SUBSURF')
            context.object.modifiers.move(len(context.object.modifiers)-1, 0)

            bpy.ops.object.modifier_apply(modifier=subsurf_modifier.name)

            bpy.ops.object.mode_set(mode='EDIT')
            bpy.ops.mesh.select_all(action='DESELECT')

            verts = ob.data.vertices

            bpy.ops.object.mode_set(mode='OBJECT')
            verts[-1].select = True
            bpy.ops.object.mode_set(mode='EDIT')
            bpy.ops.mesh.select_more(use_face_step=False)

            bpy.ops.mesh.select_similar(
                type='EDGE', compare='EQUAL', threshold=0.01)
            bpy.ops.mesh.select_all(action='INVERT')

            bpy.ops.mesh.dissolve_verts()
            bpy.ops.mesh.select_all(action='DESELECT')

            bpy.ops.mesh.select_non_manifold(
                extend=False, use_wire=False, use_boundary=True,
                use_multi_face=False, use_non_contiguous=False, use_verts=False)
            bpy.ops.mesh.select_more()

            # find boundaries
            bpy.ops.object.mode_set(mode='OBJECT')
            bound_v = [v.index for v in ob.data.vertices if v.select]
            bound_e = [e.index for e in ob.data.edges if e.select]
            bound_p = [p.index for p in ob.data.polygons if p.select]
            bpy.ops.object.mode_set(mode='EDIT')

            # select quad faces
            context.tool_settings.mesh_select_mode = (False, False, True)
            bpy.ops.mesh.select_face_by_sides(number=4, extend=False)

            # deselect boundaries
            bpy.ops.object.mode_set(mode='OBJECT')
            for i in bound_v:
                context.active_object.data.vertices[i].select = False
            for i in bound_e:
                context.active_object.data.edges[i].select = False
            for i in bound_p:
                context.active_object.data.polygons[i].select = False

            bpy.ops.object.mode_set(mode='EDIT')

            context.tool_settings.mesh_select_mode = (False, False, True)
            bpy.ops.mesh.edge_face_add()
            context.tool_settings.mesh_select_mode = (True, False, False)
            bpy.ops.mesh.select_all(action='DESELECT')

            # delete boundaries
            bpy.ops.mesh.select_non_manifold(
                    extend=False, use_wire=True, use_boundary=True,
                    use_multi_face=False, use_non_contiguous=False, use_verts=True
                    )
            bpy.ops.mesh.delete(type='VERT')

            # remove middle vertices
            bm = bmesh.from_edit_mesh(ob.data)
            for v in bm.verts:
                if len(v.link_edges) == 2 and len(v.link_faces) < 3:
                    v.select = True

            # dissolve
            bpy.ops.mesh.dissolve_verts()
            bpy.ops.mesh.select_all(action='DESELECT')

            # remove border faces
            if not self.preserve_borders:
                bpy.ops.mesh.select_non_manifold(
                    extend=False, use_wire=False, use_boundary=True,
                    use_multi_face=False, use_non_contiguous=False, use_verts=False
                    )
                bpy.ops.mesh.select_more()
                bpy.ops.mesh.delete(type='FACE')

            # clean wires
            bpy.ops.mesh.select_non_manifold(
                    extend=False, use_wire=True, use_boundary=False,
                    use_multi_face=False, use_non_contiguous=False, use_verts=False
                    )
            bpy.ops.mesh.delete(type='EDGE')

            bpy.ops.object.mode_set(mode='OBJECT')
            ob0.data.name = mesh_name
            doneMeshes.append(mesh_name)

            for o in clones:
                o.data = ob.data
            bm.free()

        for o in sel:
            o.select_set(True)

        context.view_layer.objects.active = act
        bpy.ops.object.mode_set(mode=mode)

        return {'FINISHED'}



# SPDX-FileCopyrightText: 2002-2022 Blender Authors
#
# SPDX-License-Identifier: GPL-2.0-or-later

# Add directory with module to the path.
import sys
sys.path.append(sys.argv[1])

# Just import bpy and see if there are any dynamic loader errors.
import bpy



# SPDX-FileCopyrightText: 2022 Blender Foundation
#
# SPDX-License-Identifier: GPL-2.0-or-later

#                                                                              #
#                          (c)  Alessandro Zomparelli                          #
#                                     (2020)                                   #
#                                                                              #
# http://www.co-de-it.com/                                                     #
#                                                                              #
################################################################################

import bpy
import numpy as np

import colorsys
from numpy import *

from bpy.types import (
        Operator,
        Panel
        )

from bpy.props import (
    BoolProperty,
    EnumProperty,
    FloatProperty,
    IntProperty,
    StringProperty,
    FloatVectorProperty,
    IntVectorProperty
)

from .utils import *

class random_materials(Operator):
    bl_idname = "object.random_materials"
    bl_label = "Random Materials"
    bl_description = "Assign random materials to the faces of the mesh"
    bl_options = {'REGISTER', 'UNDO'}

    prefix : StringProperty(
        name="Prefix", default="Random.", description="Name prefix")

    color_A : FloatVectorProperty(name="Color A",
                                    subtype='COLOR_GAMMA',
                                    min=0,
                                    max=1,
                                    default=[0,0,0])

    color_B : FloatVectorProperty(name="Color B",
                                    subtype='COLOR_GAMMA',
                                    min=0,
                                    max=1,
                                    default=[1,1,1])

    hue : FloatProperty(name="Hue", min=0, max=1, default=0.5)
    hue_variation : FloatProperty(name="Hue Variation", min=0, max=1, default=0.6)

    seed : IntProperty(
        name="Seed", default=0, description="Random seed")

    count : IntProperty(
        name="Count", default=3, min=2, description="Count of random materials")

    generate_materials : BoolProperty(
        name="Generate Materials", default=False, description="Automatically generates new materials")

    random_colors : BoolProperty(
        name="Random Colors", default=True, description="Colors are automatically generated")

    executed = False

    @classmethod
    def poll(cls, context):
        try: return context.object.type == 'MESH'
        except: return False

    def draw(self, context):
        layout = self.layout
        col = layout.column(align=True)
        col.prop(self, "seed")
        col.prop(self, "generate_materials")
        if self.generate_materials:
            col.prop(self, "prefix")
            col.separator()
            col.prop(self, "count")
            #row = col.row(align=True)
            col.separator()
            col.label(text='Colors:')
            col.prop(self, "hue")
            col.prop(self, "hue_variation")
            #col.prop(self, "random_colors")
            if not self.random_colors:
                col.prop(self, "color_A")
                col.prop(self, "color_B")

    def execute(self, context):
        bpy.ops.object.mode_set(mode='OBJECT')
        ob = context.active_object
        if len(ob.material_slots) == 0 and not self.executed:
            self.generate_materials = True
        if self.generate_materials:
            colA = self.color_A
            colB = self.color_B
            h1 = (self.hue - self.hue_variation/2)
            h2 = (self.hue + self.hue_variation/2)
            count = self.count
            ob.data.materials.clear()
            materials = []
            for i in range(count):
                mat_name = '{}{:03d}'.format(self.prefix,i)
                mat = bpy.data.materials.new(mat_name)
                if self.random_colors:
                    mat.diffuse_color = colorsys.hsv_to_rgb((h1 + (h2-h1)/(count)*i)%1, 1, 1)[:] + (1,)
                else:
                    mat.diffuse_color = list(colA + (colB - colA)/(count-1)*i) + [1]
                ob.data.materials.append(mat)
        else:
            count = len(ob.material_slots)
        np.random.seed(seed=self.seed)
        n_faces = len(ob.data.polygons)
        if count > 0:
            rand = list(np.random.randint(count, size=n_faces))
            ob.data.polygons.foreach_set('material_index',rand)
            ob.data.update()
        self.executed = True
        return {'FINISHED'}

class weight_to_materials(Operator):
    bl_idname = "object.weight_to_materials"
    bl_label = "Weight to Materials"
    bl_description = "Assign materials to the faces of the mesh according to the active Vertex Group"
    bl_options = {'REGISTER', 'UNDO'}

    prefix : StringProperty(
        name="Prefix", default="Weight.", description="Name prefix")

    hue : FloatProperty(name="Hue", min=0, max=1, default=0.5)
    hue_variation : FloatProperty(name="Hue Variation", min=0, max=1, default=0.3)

    count : IntProperty(
        name="Count", default=3, min=2, description="Count of random materials")

    generate_materials : BoolProperty(
        name="Generate Materials", default=False, description="Automatically generates new materials")

    mode : EnumProperty(
        items=(
                ('MIN', "Min", "Use the min weight value"),
                ('MAX', "Max", "Use the max weight value"),
                ('MEAN', "Mean", "Use the mean weight value")
                ),
        default='MEAN',
        name="Mode"
        )

    vg = None

    @classmethod
    def poll(cls, context):
        try: return context.object.type == 'MESH'
        except: return False

    def draw(self, context):
        layout = self.layout
        col = layout.column(align=True)
        col.prop(self, "mode")
        col.prop(self, "generate_materials")
        if self.generate_materials:
            col.prop(self, "prefix")
            col.separator()
            col.prop(self, "count")
            #row = col.row(align=True)
            col.separator()
            col.label(text='Colors:')
            col.prop(self, "hue")
            col.prop(self, "hue_variation")

    def execute(self, context):
        ob = context.active_object
        if self.vg == None:
            self.vg = ob.vertex_groups.active_index
        vg = ob.vertex_groups[self.vg]
        if vg == None:
            self.report({'ERROR'}, "The selected object doesn't have any Vertex Group")
            return {'CANCELLED'}
        weight = get_weight_numpy(vg, len(ob.data.vertices))
        if self.generate_materials:
            h1 = (self.hue - self.hue_variation/2)
            h2 = (self.hue + self.hue_variation/2)
            count = self.count
            ob.data.materials.clear()
            materials = []
            for i in range(count):
                mat_name = '{}{:03d}'.format(self.prefix,i)
                mat = bpy.data.materials.new(mat_name)
                mat.diffuse_color = colorsys.hsv_to_rgb((h1 + (h2-h1)/(count)*i)%1, 1, 1)[:] + (1,)
                ob.data.materials.append(mat)
        else:
            count = len(ob.material_slots)

        faces_weight = []
        for p in ob.data.polygons:
            verts_id = np.array([v for v in p.vertices])
            face_weight = weight[verts_id]
            if self.mode == 'MIN': w = face_weight.min()
            if self.mode == 'MAX': w = face_weight.max()
            if self.mode == 'MEAN': w = face_weight.mean()
            faces_weight.append(w)
        faces_weight = np.array(faces_weight)
        faces_weight = faces_weight * count
        faces_weight.astype('int')
        ob.data.polygons.foreach_set('material_index',list(faces_weight))
        ob.data.update()
        bpy.ops.object.mode_set(mode='OBJECT')
        return {'FINISHED'}



# SPDX-FileCopyrightText: 2011-2022 Blender Foundation
#
# SPDX-License-Identifier: GPL-2.0-or-later


class CopyPreset(object):
    def __init__(self, ui_name, rna_enum, elements):
        self.ui_name = ui_name
        self.rna_enum = rna_enum
        self.elements = elements


presets = (CopyPreset("Resolution",
                      ("resolution", "Render Resolution", "Render resolution and aspect ratio settings"),
                      {"resolution_x", "resolution_y", "pixel_aspect_x", "pixel_aspect_y"}),
           CopyPreset("Scale",
                      ("scale", "Render Scale", "The “Render Scale” setting"),
                      {"resolution_percentage"}),
           CopyPreset("Threads",
                      ("threads", "Render Threads", "The thread mode and number settings"),
                      {"threads_mode", "threads"}),
           CopyPreset("Stamp",
                      ("stamp", "Render Stamp", "The Stamp toggle"),
                      {"use_stamp"})
           )
